@@ -15,7 +15,7 @@
 #
 import pickle
 from abc import abstractmethod
-from typing import List, Dict, Union
+from typing import List, Dict, Tuple, Optional, Union
 
 import pandas as pd
 import mlflow
@@ -26,7 +26,8 @@ from mlflow.utils.environment import _mlflow_conda_env
 
 from databricks.automl_runtime.forecast import OFFSET_ALIAS_MAP, DATE_OFFSET_KEYWORD_MAP
 from databricks.automl_runtime.forecast.model import ForecastModel, mlflow_forecast_log_model
-from databricks.automl_runtime.forecast.utils import calculate_period_differences, is_frequency_consistency
+from databricks.automl_runtime.forecast.utils import calculate_period_differences, is_frequency_consistency, \
+    make_future_dataframe, make_single_future_dataframe
 
 
 ARIMA_CONDA_ENV = _mlflow_conda_env(
@@ -86,7 +87,8 @@ class ArimaModel(AbstractArimaModel):
     """
 
     def __init__(self, pickled_model: bytes, horizon: int, frequency: str,
-                 start_ds: pd.Timestamp, end_ds: pd.Timestamp, time_col: str) -> None:
+                 start_ds: pd.Timestamp, end_ds: pd.Timestamp,
+                 time_col: str, exogenous_cols: Optional[List[str]] = None) -> None:
         """
         Initialize the mlflow Python model wrapper for ARIMA.
         :param pickled_model: the pickled ARIMA model as a bytes object.
@@ -95,6 +97,8 @@ class ArimaModel(AbstractArimaModel):
         :param start_ds: the start time of training data
         :param end_ds: the end time of training data
         :param time_col: the column name of the time column
+        :param exogenous_cols: Optional list of column names of exogenous variables. If provided, these columns are
+        used as additional features in arima model.
         """
         super().__init__()
         self._pickled_model = pickled_model
@@ -103,6 +107,7 @@ class ArimaModel(AbstractArimaModel):
         self._start_ds = pd.to_datetime(start_ds)
         self._end_ds = pd.to_datetime(end_ds)
         self._time_col = time_col
+        self._exogenous_cols = exogenous_cols
 
     def model(self) -> pmdarima.arima.ARIMA:
         """
@@ -111,22 +116,48 @@ class ArimaModel(AbstractArimaModel):
         """
         return pickle.loads(self._pickled_model)
 
-    def predict_timeseries(self, horizon: int = None, include_history: bool = True) -> pd.DataFrame:
+    def predict_timeseries(
+        self,
+        horizon: int = None,
+        include_history: bool = True,
+        df: Optional[pd.DataFrame] = None) -> pd.DataFrame:
         """
         Predict target column for given horizon_timedelta and history data.
         :param horizon: int number of periods to forecast forward.
         :param include_history: Boolean to include the historical dates in the data
             frame for predictions.
+        :param df: A pd.Dataframe containing regressors (exogenous variables), if they were used to train the model.
         :return: A pd.DataFrame with the forecasts and confidence intervals for given horizon_timedelta and history data.
         """
         horizon = horizon or self._horizon
-        future_pd = self._forecast(horizon)
+        X = None
+        if self._exogenous_cols and df is not None:
+            time_col = self._time_col if self._time_col in df.columns else "ds"
+            X = (df[df[time_col] > self._end_ds].set_index(time_col))[self._exogenous_cols]
+        future_pd = self._forecast(horizon, X)
         if include_history:
-            in_sample_pd = self._predict_in_sample()
+            in_sample_pd = self._predict_in_sample(start_ds=self._start_ds, end_ds=self._end_ds, X=X)
             return pd.concat([in_sample_pd, future_pd]).reset_index(drop=True)
         else:
             return future_pd
 
+    def make_future_dataframe(self, horizon: int = None, include_history: bool = True) -> pd.DataFrame:
+        """
+        Generate future dataframe by calling the API from prophet
+        :param horizon: Int number of periods to forecast forward.
+        :param include_history: Boolean to include the historical dates in the data
+            frame for predictions.
+        :return: pd.Dataframe that extends forward from the end of self.history for the
+        requested number of periods.
+        """
+        return make_single_future_dataframe(
+            start_time=self._start_ds,
+            end_time=self._end_ds,
+            horizon=horizon or self._horizon,
+            frequency=self._frequency,
+            include_history=include_history
+        )
+
     def predict(self, context: mlflow.pyfunc.model.PythonModelContext, model_input: pd.DataFrame) -> pd.Series:
         """
         Predict API from mlflow.pyfunc.PythonModel.
@@ -171,18 +202,29 @@ class ArimaModel(AbstractArimaModel):
         # Out-of-sample prediction if needed
         horizon = calculate_period_differences(self._end_ds, max(df["ds"]), self._frequency)
         if horizon > 0:
-            future_pd = self._forecast(horizon)
+            X_future = df[df["ds"] > self._end_ds].set_index("ds")
+            future_pd = self._forecast(
+                horizon,
+                X=X_future[self._exogenous_cols] if self._exogenous_cols else None)
             preds_pds.append(future_pd)
         # In-sample prediction if needed
         if pred_start_ds <= self._end_ds:
-            in_sample_pd = self._predict_in_sample(start_ds=pred_start_ds, end_ds=self._end_ds)
+            X_in_sample = df[df["ds"] <= self._end_ds].set_index("ds")
+            in_sample_pd = self._predict_in_sample(
+                start_ds=pred_start_ds,
+                end_ds=self._end_ds,
+                X=X_in_sample[self._exogenous_cols] if self._exogenous_cols else None)
             preds_pds.append(in_sample_pd)
         # Map predictions back to given timestamps
         preds_pd = pd.concat(preds_pds).set_index("ds")
         df = df.set_index("ds").join(preds_pd, how="left").reset_index()
         return df
 
-    def _predict_in_sample(self, start_ds: pd.Timestamp = None, end_ds: pd.Timestamp = None) -> pd.DataFrame:
+    def _predict_in_sample(
+        self,
+        start_ds: pd.Timestamp = None,
+        end_ds: pd.Timestamp = None,
+        X: pd.DataFrame = None) -> pd.DataFrame:
         if start_ds and end_ds:
             start_idx = calculate_period_differences(self._start_ds, start_ds, self._frequency)
             end_idx = calculate_period_differences(self._start_ds, end_ds, self._frequency)
@@ -190,17 +232,28 @@ class ArimaModel(AbstractArimaModel):
             start_ds = self._start_ds
             end_ds = self._end_ds
             start_idx, end_idx = None, None
+        d = self.model().order[1]
+        start_idx = max(start_idx, d)
         preds_in_sample, conf_in_sample = self.model().predict_in_sample(
-            start=start_idx, end=end_idx, return_conf_int=True)
-        periods = calculate_period_differences(start_ds, end_ds, self._frequency) + 1
-        ds_indices = self._get_ds_indices(start_ds=start_ds, periods=periods, frequency=self._frequency)
+            X=X,
+            start=start_idx,
+            end=end_idx,
+            return_conf_int=True)
+        periods = calculate_period_differences(self._start_ds, end_ds, self._frequency) + 1
+        ds_indices = self._get_ds_indices(start_ds=self._start_ds, periods=periods, frequency=self._frequency)[start_idx:]
         in_sample_pd = pd.DataFrame({'ds': ds_indices, 'yhat': preds_in_sample})
         in_sample_pd[["yhat_lower", "yhat_upper"]] = conf_in_sample
         return in_sample_pd
 
-    def _forecast(self, horizon: int = None) -> pd.DataFrame:
+    def _forecast(
+        self,
+        horizon: int = None,
+        X: pd.DataFrame = None) -> pd.DataFrame:
         horizon = horizon or self._horizon
-        preds, conf = self.model().predict(horizon, return_conf_int=True)
+        preds, conf = self.model().predict(
+            horizon,
+            X=X,
+            return_conf_int=True)
         ds_indices = self._get_ds_indices(start_ds=self._end_ds, periods=horizon + 1, frequency=self._frequency)[1:]
         preds_pd = pd.DataFrame({'ds': ds_indices, 'yhat': preds})
         preds_pd[["yhat_lower", "yhat_upper"]] = conf
@@ -212,18 +265,20 @@ class MultiSeriesArimaModel(AbstractArimaModel):
     ARIMA mlflow model wrapper for multivariate forecasting.
     """
 
-    def __init__(self, pickled_model_dict: Dict[str, bytes], horizon: int, frequency: str,
-                 start_ds_dict: Dict[str, pd.Timestamp], end_ds_dict: Dict[str, pd.Timestamp],
-                 time_col: str, id_cols: List[str]) -> None:
+    def __init__(self, pickled_model_dict: Dict[Tuple, bytes], horizon: int, frequency: str,
+                 start_ds_dict: Dict[Tuple, pd.Timestamp], end_ds_dict: Dict[Tuple, pd.Timestamp],
+                 time_col: str, id_cols: List[str], exogenous_cols: Optional[List[str]] = None) -> None:
         """
         Initialize the mlflow Python model wrapper for multiseries ARIMA.
         :param pickled_model_dict: the dictionary of binarized ARIMA models for different time series.
         :param horizon: int number of periods to forecast forward.
         :param frequency: the frequency of the time series
         :param start_ds_dict: the dictionary of the starting time of each time series in training data.
-        :param end_ds_dict: the dictionary of the starting time of each time series  in training data.
+        :param end_ds_dict: the dictionary of the end time of each time series in training data.
         :param time_col: the column name of the time column
         :param id_cols: the column names of the identity columns for multi-series time series
+        :param exogenous_cols: Optional list of column names of exogenous variables. If provided, these columns are
+        used as additional features in arima model.
         """
         super().__init__()
         self._pickled_models = pickled_model_dict
@@ -233,8 +288,9 @@ class MultiSeriesArimaModel(AbstractArimaModel):
         self._ends = end_ds_dict
         self._time_col = time_col
         self._id_cols = id_cols
+        self._exogenous_cols = exogenous_cols
 
-    def model(self, id_: str) -> pmdarima.arima.ARIMA:
+    def model(self, id_: Tuple) -> pmdarima.arima.ARIMA:
         """
         Deserialize the ARIMA model for specified time series by pickle.
         :param: id for specified time series.
@@ -242,24 +298,70 @@ class MultiSeriesArimaModel(AbstractArimaModel):
         """
         return pickle.loads(self._pickled_models[id_])
 
-    def predict_timeseries(self, horizon: int = None, include_history: bool = True) -> pd.DataFrame:
+    def make_future_dataframe(
+            self,
+            horizon: Optional[int] = None,
+            include_history: bool = True,
+            groups: List[Tuple] = None,
+    ) -> pd.DataFrame:
+        """
+        Generate dataframe with future timestamps for all valid identities
+        :param horizon: Int number of periods in the future
+        :param frequency: frequency of the history time series. It should be valid frequency
+            for pd.date_range, such as "D" or "M"
+        :param groups: the collection of group(s) to generate forecast predictions.
+            The group definiteions must be the key value
+        :return: pd.DataFrame that extends forward from history_last_date
+        """
+        if horizon is None:
+            horizon = self._horizon
+        if groups is not None:
+            model_keys = set(self._pickled_models.keys())
+            if not set(groups).issubset(model_keys):
+                raise ValueError(f"Invalid groups: {set(groups) - model_keys}.")
+        else:
+            groups = list(self._pickled_models.keys())
+
+        future_df = make_future_dataframe(
+            start_time=self._starts,
+            end_time=self._ends,
+            horizon=horizon,
+            frequency=self._frequency,
+            include_history=include_history,
+            groups=groups,
+            identity_column_names=self._id_cols
+        )
+        return future_df
+
+    def predict_timeseries(
+        self,
+        horizon: int = None,
+        include_history: bool = True,
+        df: Optional[pd.DataFrame] = None) -> pd.DataFrame:
         """
         Predict target column for given horizon_timedelta and history data.
         :param horizon: Int number of periods to forecast forward.
         :param include_history: Boolean to include the historical dates in the data
             frame for predictions.
+        :param df: A pd.Dataframe containing regressors (exogenous variables), if they were used to train the model.
         :return: A pd.DataFrame with the forecast components.
         """
         horizon = horizon or self._horizon
         ids = self._pickled_models.keys()
-        preds_dfs = list(map(lambda id_: self._predict_timeseries_single_id(id_, horizon, include_history), ids))
+        preds_dfs = list(map(lambda id_: self._predict_timeseries_single_id(id_, horizon, include_history, df), ids))
         return pd.concat(preds_dfs).reset_index(drop=True)
 
-    def _predict_timeseries_single_id(self, id_: str, horizon: int, include_history: bool = True) -> pd.DataFrame:
+    def _predict_timeseries_single_id(
+        self,
+        id_: Tuple,
+        horizon: int,
+        include_history: bool = True,
+        df: Optional[pd.DataFrame] = None) -> pd.DataFrame:
         arima_model_single_id = ArimaModel(self._pickled_models[id_], self._horizon, self._frequency,
-                                           self._starts[id_], self._ends[id_], self._time_col)
-        preds_df = arima_model_single_id.predict_timeseries(horizon, include_history)
-        preds_df["ts_id"] = id_
+                                           self._starts[id_], self._ends[id_], self._time_col, self._exogenous_cols)
+        preds_df = arima_model_single_id.predict_timeseries(horizon, include_history, df)
+        for id, col_name in zip(id_, self._id_cols):
+            preds_df[col_name] = id
         return preds_df
 
     def predict(self, context: mlflow.pyfunc.model.PythonModelContext, model_input: pd.DataFrame) -> pd.Series:
@@ -276,7 +378,7 @@ class MultiSeriesArimaModel(AbstractArimaModel):
         """
         self._validate_cols(model_input, self._id_cols + [self._time_col])
         df = model_input.copy()
-        df["ts_id"] = df[self._id_cols].apply(lambda r: "-".join(r.values.astype(str)), axis=1)
+        df["ts_id"] = df[self._id_cols].apply(tuple, axis=1)
         known_ids = set(self._pickled_models.keys())
         ids = set(df["ts_id"].unique())
         if not ids.issubset(known_ids):
@@ -288,15 +390,20 @@ class MultiSeriesArimaModel(AbstractArimaModel):
                 ),
                 error_code=INVALID_PARAMETER_VALUE,
             )
-        preds_df = df.groupby("ts_id").apply(self._predict_single_id).reset_index(drop=True)
-        df = df.merge(preds_df, how="left", on=[self._time_col, "ts_id"])  # merge predictions to original order
+        preds_df = df.groupby(self._id_cols).apply(self._predict_single_id).reset_index(drop=True)
+        df = df.merge(preds_df, how="left", on=[self._time_col] + self._id_cols)  # merge predictions to original order
         return df["yhat"]
 
     def _predict_single_id(self, df: pd.DataFrame) -> pd.DataFrame:
         id_ = df["ts_id"].to_list()[0]
-        arima_model_single_id = ArimaModel(self._pickled_models[id_], self._horizon, self._frequency,
-                                           self._starts[id_], self._ends[id_], self._time_col)
-        df["yhat"] = arima_model_single_id.predict(None, df).to_list()
+        arima_model_single_id = ArimaModel(self._pickled_models[id_],
+                                           self._horizon,
+                                           self._frequency,
+                                           self._starts[id_],
+                                           self._ends[id_],
+                                           self._time_col,
+                                           self._exogenous_cols)
+        df["yhat"] = arima_model_single_id.predict(context=None, model_input=df).to_list()
         return df
 
 
