@@ -14,7 +14,7 @@
 # limitations under the License.
 #
 import logging
-from typing import List, Optional
+from typing import Dict, List, Optional, Tuple, Union
 from databricks.automl_runtime.forecast import DATE_OFFSET_KEYWORD_MAP,\
     QUATERLY_OFFSET_ALIAS, NON_DAILY_OFFSET_ALIAS, OFFSET_ALIAS_MAP, PERIOD_ALIAS_MAP
 
@@ -22,6 +22,80 @@ import pandas as pd
 
 _logger = logging.getLogger(__name__)
 
+def make_future_dataframe(
+        start_time: Union[pd.Timestamp, Dict[Tuple, pd.Timestamp]],
+        end_time: Union[pd.Timestamp, Dict[Tuple, pd.Timestamp]],
+        horizon: int,
+        frequency: str,
+        include_history: bool = True,
+        groups: List[Tuple] = None,
+        identity_column_names: List[str] = None,
+) -> pd.DataFrame:
+    """
+    Utility function to generate the dataframe with future timestamps.
+    :param start_time: the dictionary of the starting time of each time series in training data.
+    :param end_time: the dictionary of the end time of each time series in training data.
+    :param horizon: int number of periods to forecast forward.
+    :param frequency: the frequency of the time series
+    :param include_history:
+    :param groups: the collection of group(s) to generate forecast predictions.
+    :param identity_column_names: Column names of the identity columns
+    :return: pd.DataFrame that extends forward
+    """
+    if groups is None:
+        return make_single_future_dataframe(start_time, end_time, horizon, frequency)
+
+    future_df_list = []
+    for group in groups:
+        if type(start_time) is dict:
+            group_start_time = start_time[group]
+        else:
+            group_start_time = start_time
+        if type(end_time) is dict:
+            group_end_time = end_time[group]
+        else:
+            group_end_time = end_time
+        df = make_single_future_dataframe(group_start_time, group_end_time, horizon, frequency, include_history)
+        for idx, identity_column_name in enumerate(identity_column_names):
+            df[identity_column_name] = group[idx]
+        future_df_list.append(df)
+    return pd.concat(future_df_list)
+
+def make_single_future_dataframe(
+        start_time: pd.Timestamp,
+        end_time: pd.Timestamp,
+        horizon: int,
+        frequency: str,
+        include_history: bool = True,
+        column_name: str = "ds"
+) -> pd.DataFrame:
+    """
+    Generate future dataframe for one model
+    :param start_time: The starting time of time series of the training data.
+    :param end_time: The end time of time series of the training data.
+    :param horizon: Int number of periods to forecast forward.
+    :param frequency: The frequency of the time series
+    :param include_history: Boolean to include the historical dates in the data
+            frame for predictions.
+    :param column_name: column name of the time column. Default is "ds".
+    :return:
+    """
+    offset_freq = DATE_OFFSET_KEYWORD_MAP[OFFSET_ALIAS_MAP[frequency]]
+    unit_offset = pd.DateOffset(**offset_freq)
+    end_time = pd.Timestamp(end_time)
+
+    if include_history:
+        start_time = start_time
+    else:
+        start_time = end_time + unit_offset
+
+    date_rng = pd.date_range(
+        start=start_time,
+        end=end_time + unit_offset*horizon,
+        freq=unit_offset
+    )
+    return pd.DataFrame(date_rng, columns=[column_name])
+
 def get_validation_horizon(df: pd.DataFrame, horizon: int, unit: str) -> int:
     """
     Return validation_horizon, which is the lesser of `horizon` and one quarter of the dataframe's timedelta
