@@ -16,8 +16,14 @@
 
 from databricks.automl_runtime.sklearn.date_transformer import DateTransformer
 from databricks.automl_runtime.sklearn.datetime_imputer import DatetimeImputer
+from databricks.automl_runtime.sklearn.interpolation_imputer import InterpolationImputer
+from databricks.automl_runtime.sklearn.pandas_column_transformer import PandasColumnTransformer
+from databricks.automl_runtime.sklearn.pandas_simple_imputer import PandasSimpleImputer
+from databricks.automl_runtime.sklearn.pandas_transformer_wrapper import PandasTransformerWrapper
 from databricks.automl_runtime.sklearn.timestamp_transformer import TimestampTransformer
 from databricks.automl_runtime.sklearn.transformed_target_classifier import TransformedTargetClassifier
 from databricks.automl_runtime.sklearn.one_hot_encoder import OneHotEncoder
 
-__all__ = ["DateTransformer", "DatetimeImputer", "TimestampTransformer", "TransformedTargetClassifier", "OneHotEncoder"]
+__all__ = ["DateTransformer", "DatetimeImputer", "InterpolationImputer", "PandasColumnTransformer",
+           "PandasSimpleImputer", "PandasTransformerWrapper", "TimestampTransformer", "TransformedTargetClassifier",
+           "OneHotEncoder"]
